namespace NXyn.Aux
    open System
    open System.Collections
    open System.Collections.Generic
    
    module ConvertMisc =
        let String_of_BitArray(b:BitArray) =
            let len = b.Length
            let arr = Array.create len '0'
            for i = 0 to len - 1 do if b.[i] then arr.[len - 1 - i] <- '1'
            new string(arr)

        let BitArray_of_String(s:string) =    
            let len = s.Length
            let d = new System.Collections.BitArray(len)
            for i = 0 to len - 1 do
                match s.[i] with
                    | '1' -> d.[len - 1 - i] <- true
                    | '0' -> ()
                    | chr -> failwithf "JSON.BitArray_of_string: invalid character '%c'" chr
            d


        let bool_of_string(s) = Boolean.Parse(s)

        (* similar to System.Int32.Parse, but more flexible (radices supported in 2..36) and avoids construction of temporary string. *)

        (* TBD is there something similar in BigIntConverter? => merge, factor *)
        let value_of_digit (r:int) =
            if r <= 1 || r > 36 then failwithf "value_of_digit: invalid radix %d" r
            let tbl = Array.create 256 (-1)
            for i = 0 to min 9 (r-1) do tbl.[i + int '0'] <- i
            for i = 0 to (r-11) do tbl.[i + int 'A'] <- i+10; tbl.[i + int 'a'] <- i+10
            fun (digit:char) ->
                let v = if tbl.Length <= int digit then -1 else tbl.[int digit]
                if v >= 0 then v else failwithf "value_of_digit(%d): invalid digit '%c'" r digit
 
        let valueOfHexdigit(hexdigit:char) = value_of_digit 16 hexdigit


        (* case-insensitive alphanumeric numeral system using ASCII characters *)
        type BigIntConverter(radix:int) = 
            let _ = assert (radix >= 2 && radix <= 36)
            let ValueOfCharCode = Array.create 256 (-1)
            let _ = for i = 0 to min 9 radix  do ValueOfCharCode.[48 + i] <- i
            let _ = for i = 0 to (radix - 11) do ValueOfCharCode.[65 + i] <- 10 + i; ValueOfCharCode.[97 + i] <- 10 + i
            let CharOfValue = Array.init radix (fun i -> Char.ToLower(char (Array.findIndex ((=)i) ValueOfCharCode)))
            let Base = bigint radix

            member x.TryParse (s:string) =
                let rec loop (v:bigint) (i:int) =
                    if i >= s.Length then Some v  
                    elif ValueOfCharCode.[int s.[i]] < 0 then None
                    else loop (v * Base + bigint ValueOfCharCode.[int s.[i]]) (i+1)
                if s.Length = 0 then None
                                else if s.[0] <> '-' then loop 0I 0
                                                     else if s.Length = 1 then None
                                                                          else match loop 0I 1 with
                                                                                   | Some i -> Some(-i)
                                                                                   | None -> None
            member x.Parse (s:string) =
                match x.TryParse s with
                    | Some i -> i
                    | None -> failwithf "ConvertMisc.BigIntConverter.Parse: cannot parse %s as base-%d integer" s radix

            member x.Unparse (v:bigint) =
              if v = 0I then "0"
              else
                let isNegative = v < 0I
                let arr = new ResizeArray<char>()

                let rec loop (v:bigint) =
                    if v = 0I then 
                        if isNegative then arr.Add('-')
                        arr.Reverse ()
                        new String(arr.ToArray ())
                    else
                        arr.Add(CharOfValue.[int (v % Base)])
                        loop(v/Base)
                loop(abs(v))

        let base2bigintconv  = BigIntConverter(2)
        let base8bigintconv  = BigIntConverter(8)
        let base10bigintconv = BigIntConverter(10)
        let base16bigintconv = BigIntConverter(16)

