namespace NXyn.EmitFSharp

  open System
  open System.Collections.Generic

  open System.Reflection
  open Microsoft.FSharp.Reflection
  
  open NXyn.EmitFSharp.Common

  module Fold =
      let GetUnionCasesAndFields (t:Type) = Array.map (fun (uc:UnionCaseInfo) -> let fs = uc.GetFields() in (uc,fs)) (FSharpType.GetUnionCases(t))

      let use_CatchAll = true

      let OutputCode (name:string) newTypes (t:Type) (mtype:Type) (userFilter:PropertyInfo -> bool) (folderOption:string option) (stateOption:string option) (w:IO.TextWriter) =
          match RemoveIdenticalTypes mtype userFilter newTypes with
              | [] ->
                  match (folderOption,stateOption) with
                      | (Some f, Some s) -> fprintfn w "let %s arg = %s"          name s
                      | (Some f, None)   -> fprintfn w "let %s s arg = s"         name
                      | (None,   Some s) -> fprintfn w "let %s fold_it arg = %s"  name s
                      | (None,   None)   -> fprintfn w "let %s fold_it s arg = s" name
              | newTypes ->
                  let map = new Dictionary<Type,string>()
                  List.iteri (fun (index:int) (t:Type) -> map.Add(t, "fold_" + string index)) newTypes
                  let fold_it = match folderOption with Some s -> s | None -> "fold_it"

                  let writeFold_of_argument (i:int) (f:PropertyInfo) =
                      if   f.PropertyType = mtype && userFilter f then fprintf w "let s = %s s a%d in " fold_it              i
                      elif map.ContainsKey f.PropertyType         then fprintf w "let s = %s s a%d in " map.[f.PropertyType] i

                  let writeFold_of_type (t:Type) =
                      (* look at JSON.EmitIO how to do this better... (more complete) *)
                      if   is_FSharpList(t)      then fprintfn w "%s s xs = (* F# list *) List.fold %s s xs"     map.[t] map.[firstGenericArgument_of_Type(t)]
                      elif is_FSharpOption(t)    then fprintfn w "%s s xs = (* F# option *) Option.fold %s s xs" map.[t] map.[firstGenericArgument_of_Type(t)]
                      elif FSharpType.IsRecord t then
                          let fs = FSharpType.GetRecordFields(t)
                          WriteList w (map.[t] + " s ( { ") "; " (" } : " + (FSharpTypeName t) + " ) = ") fs.Length (fun i -> fs.[i].Name + "=a" + string i)
                          Array.iteri writeFold_of_argument fs
                          fprintfn w "s"
                      elif FSharpType.IsUnion t then
                          fprintfn w "%s s = function" map.[t]
                          let t_name = if t.Name = "FSharpOption`1" then "Option" else t.Name (* HACK! DUP! EmitFSharpMap *)
                          let some_notcovered = ref false
                          for ((uc,fs):(UnionCaseInfo * PropertyInfo [])) in GetUnionCasesAndFields(t) do
                              if Array.exists (fun(f:PropertyInfo) -> map.ContainsKey f.PropertyType (* BOILER *)
                                                                      || f.PropertyType = mtype && userFilter f) fs then
                                  WriteList w ("    | " + t_name + "." + uc.Name + "(") "," ") -> " fs.Length (fun i -> "a" + string i)
                                  Array.iteri writeFold_of_argument fs
                                  fprintfn w "s"
                              elif use_CatchAll then some_notcovered := true 
                              elif fs = [||]    then fprintfn w   "    | %s.%s -> s"   t_name uc.Name
                                                else fprintfn w   "    | %s.%s _ -> s" t_name uc.Name
                          if !some_notcovered then fprintfn w "    | _ -> s"
                      else fprintfn w "%s s (arg:%s) = %s s arg" map.[t] (FSharpTypeName(t)) fold_it

                  fprintfn w "let %s%s =" name (if Option.isNone(folderOption) then " fold_it" else "")

                  List.iteri (fun (i:int) (t:Type) -> fprintf w "  %s " (if i = 0 then "let rec" else "and"); writeFold_of_type t) newTypes

                  match stateOption with
                      | None    -> fprintfn w "  fun s arg -> %s s arg" map.[t]
                      | Some(s) -> fprintfn w "  fun arg -> %s %s arg"  map.[t] s
          fprintfn w "\n"
