
module MoreAstIO

open System
open System.Collections.Generic

open X64instr
open Ast

open NXyn.JSON

open InstrIO
open MoreInstrIO

let ILine_of_ELine (line:ELine) =
                   match line
                    with | Label(lbl) -> Some(ALabel(lbl,1))
                         | Tagged(d,args) when String.IsNullOrEmpty(d) -> None
                         | Tagged(d,args) when d.[0] = '.' -> Some(ILine.MakeDirective (d.Substring(1)) args)
                         | Tagged(d,args) -> match Instr_of_ELine(d,args)
                                              with | Some(i) -> Some(Instr(i))
                                                   | None -> match line.AddImpliedLiteral
                                                              with | Label(lbl) -> Some(ALabel(lbl,1))
                                                                   | Tagged(op,args) -> Instr_of_ELine(op,args) |> Option.map ILine.Instr
//                                                                   | Tagged(op,args) -> match Instr_of_ELine(op,args) // Option.
//                                                                                         with | Some(i) -> Some(Instr(i))
//                                                                                              | None -> None


let rec WriteILine_to (line:ILine) (w:IO.TextWriter) =
                      match line
                       with | ALabel(lbl,alignment) -> if alignment > 1 then WriteILine_to (Directive("balign", [Item.Iexpr(Iexpr.Int(bigint alignment))])) w
                                                                             fprintfn w ""
                                                       fprintf w "%s:" lbl
                            | Directive(t,args) -> fprintf w "\t.%s" t
                                                   printAsmIndent w (t.Length + 1)
                                                   Write_ASTlineargs w args
                            | Instr(i) -> WriteInstr_to i w


let String_of_ILine (line:ILine) =
                    let w = new IO.StringWriter()
                    WriteILine_to line w
                    w.ToString()


let TryParseASTitem(s) = try         Some(Microsoft.FSharp.Text.Lexing.LexBuffer<char>.FromString(s) |> AttParser.ASTitem AttLexer.token)
                         with exn -> None

let TryParseASTline(s) = try         Some(Microsoft.FSharp.Text.Lexing.LexBuffer<char>.FromString(s + Environment.NewLine) |> AttParser.ASTline AttLexer.token)
                         with exn -> None

let (|CompactString|_|) x = match x with Item.Iexpr(e) -> let s = e.CompactString in if TryParseASTitem(s) = Some(x) then Some(s) else None
                                       | _ -> None

