
module MoreInstrIO

open System
open System.Collections.Generic

open X64instr
open InstrIO


let Ms_of_Instr(i:Instr) = List.map (fun (M8.S(m))   -> (1,m))  (M8s_of_Instr(i))   @
                           List.map (fun (M16.S(m))  -> (2,m))  (M16s_of_Instr(i))  @
                           List.map (fun (M32.S(m))  -> (4,m))  (M32s_of_Instr(i))  @
                           List.map (fun (M64.S(m))  -> (8,m))  (M64s_of_Instr(i))  @
                           List.map (fun (M128.S(m)) -> (16,m)) (M128s_of_Instr(i)) @
                           List.map (fun (M256.S(m)) -> (32,m)) (M256s_of_Instr(i))
                           
let Mds_of_Instr(i:Instr) = List.map (fun (M8.S(m))   -> (1,m)) (M8ds_of_Instr(i))    @
                            List.map (fun (M16.S(m))  -> (2,m)) (M16ds_of_Instr(i))   @
                            List.map (fun (M32.S(m))  -> (4,m)) (M32ds_of_Instr(i))   @
                            List.map (fun (M64.S(m))  -> (8,m)) (M64ds_of_Instr(i))   @
                            List.map (fun (M128.S(m)) -> (16,m)) (M128ds_of_Instr(i)) @
                            List.map (fun (M256.S(m)) -> (32,m)) (M256ds_of_Instr(i))

let Md_of_Instr(i:Instr) = List.map (fun (M8.S(m))   -> (1,m)) (M8d_of_Instr(i))    @
                           List.map (fun (M16.S(m))  -> (2,m)) (M16d_of_Instr(i))   @
                           List.map (fun (M32.S(m))  -> (4,m)) (M32d_of_Instr(i))   @
                           List.map (fun (M64.S(m))  -> (8,m)) (M64d_of_Instr(i))   @
                           List.map (fun (M128.S(m)) -> (16,m)) (M128d_of_Instr(i)) @
                           List.map (fun (M256.S(m)) -> (32,m)) (M256d_of_Instr(i))

let FixedX_of_Instr(i:Instr) = FixedXd_of_Instr(i) @ FixedXds_of_Instr(i) @ FixedXs_of_Instr(i)

let AllRs_of_Instr(i:Instr)  = FixedRs_of_Instr(i)  @ Rs_of_Instr(i)
let AllRds_of_Instr(i:Instr) = List.filter (fun(r) -> size_of_R(r) <  4) (FixedRd_of_Instr(i) @ Rd_of_Instr(i)) @ FixedRds_of_Instr(i) @ Rds_of_Instr(i)
let AllRd_of_Instr(i:Instr)  = List.filter (fun(r) -> size_of_R(r) >= 4) (FixedRd_of_Instr(i) @ Rd_of_Instr(i))
let AllR_of_Instr(i:Instr)   = AllRd_of_Instr(i) @ AllRds_of_Instr(i) @ AllRs_of_Instr(i)

let AllXs_of_Instr(i:Instr)  = FixedXs_of_Instr(i)  @ Xs_of_Instr(i)
let AllXds_of_Instr(i:Instr) = FixedXds_of_Instr(i) @ Xds_of_Instr(i)
let AllXd_of_Instr(i:Instr)  = FixedXd_of_Instr(i)  @ Xd_of_Instr(i)
let AllX_of_Instr(i:Instr)   = AllXd_of_Instr(i) @ AllXds_of_Instr(i) @ AllXs_of_Instr(i)

let AllYs_of_Instr(i:Instr)  = FixedYs_of_Instr(i)  @ Ys_of_Instr(i)
let AllYds_of_Instr(i:Instr) = FixedYds_of_Instr(i) @ Yds_of_Instr(i)
let AllYd_of_Instr(i:Instr)  = FixedYd_of_Instr(i)  @ Yd_of_Instr(i)
let AllY_of_Instr(i:Instr)   = AllYd_of_Instr(i) @ AllYds_of_Instr(i) @ AllYs_of_Instr(i)



let Instr_is_J(i:Instr)   = Instr_is_Jfollow(i) || Instr_is_Jthru(i)
let Instr_is_nop(i:Instr) = Instr_has_no_effect(i) || Instr_is_NoOp(i)

(***)

let D_of_Instr (i:Instr) =
                   let d = new ResizeArray<Operand>()
                   if Instr_is_J(i) then d.Add(Control)
                   if Instr_uses_FixedMem(i) then d.Add(Gmemory)
                   Md_of_Instr(i)    |> List.iter (function StackLoc(size,ofs) -> for i = 0 to size-1 do d.Add(PstackB(ofs+i))
                                                          | _ -> d.Add(Gmemory))
                   AllRd_of_Instr(i) |> List.iter (Operand_of_R >> d.Add)
                   Fd_of_Instr(i)    |> List.iter (Operand_of_F >> d.Add)
                   AllXd_of_Instr(i) |> List.iter (Operand_of_X >> d.Add)
                   AllYd_of_Instr(i) |> List.iter (Operand_of_Y >> d.Add)
//                   AllVd_of_Instr(i) |> List.iter (Operand_of_V >> d.Add)
                   d.ToArray() |> Set.ofArray



let DS_of_Instr (i:Instr) =
                    let ds = new ResizeArray<Operand>()
                    if Instr_uses_FixedMem(i) then ds.Add(Gmemory)
                    Mds_of_Instr(i)    |> List.iter (function StackLoc(size,ofs) -> for i = 0 to size-1 do ds.Add(PstackB(ofs+i))
                                                                  | _ -> ds.Add(Gmemory))
                    AllRds_of_Instr(i) |> List.iter (Operand_of_R >> ds.Add)
                    AllXds_of_Instr(i) |> List.iter (Operand_of_X >> ds.Add)
                    AllYds_of_Instr(i) |> List.iter (Operand_of_Y >> ds.Add)
//                    AllVds_of_Instr(i) |> List.iter (Operand_of_V >> ds.Add)
                    ds.ToArray() |> Set.ofArray



let S_of_Instr (kDict:Dictionary<String,Int32>) (returnsVoid:Boolean) (i:Instr) =
                   let s = new ResizeArray<Operand>()
                   s.Add(Control)
                   if Instr_uses_FixedMem(i) then s.Add(Gmemory)
                   Ms_of_Instr(i)    |> List.iter (function StackLoc(size,ofs) -> for i = 0 to size-1 do s.Add(PstackB(ofs+i))
                                                          | (size,(K(Id({Name = k})) | K32eip(Id({Name = k})) | K64rip(Id({Name = k})))) when kDict.ContainsKey(k) && kDict.[k] = size -> ()
                                                          | _ -> s.Add(Gmemory))
                   AllRs_of_Instr(i) |> List.iter (Operand_of_R >> s.Add)
                   Fs_of_Instr(i)    |> List.iter (Operand_of_F >> s.Add)
                   AllXs_of_Instr(i) |> List.iter (Operand_of_X >> s.Add)
                   AllYs_of_Instr(i) |> List.iter (Operand_of_Y >> s.Add)
//                   AllVs_of_Instr(i) |> List.iter (Operand_of_V >> s.Add)
                   if Instr_is_FunReturn(i) then (if returnsVoid then MinReturnRegisters else MaxReturnRegisters) |> List.iter s.Add
                   s.ToArray() |> Set.ofArray

(***)

let WriteInstr_to (i:Instr) (w:IO.TextWriter) = Ast.WriteELine_to (InstrIO.ELine_of_Instr(i)) w

let String_of_Instr (i:Instr) = let w = new IO.StringWriter()
                                WriteInstr_to i w
                                w.ToString()

let Map_XYto08_in_Instr (i:Instr) = i |> Map_X_in_Instr (fun(x:X) -> x.Normalize) |> Map_Y_in_Instr (fun(y:Y) -> y.Normalize)
