
module OptVecReallocGA

open System
open System.Collections.Generic

open Ast
open X64instr

open InstrIO
open MoreInstrIO

open AstIO
open MoreAstIO


open NXyn
open NXyn.Aux
open NXyn.Aux.Global

open NXyn.Aux.IntMisc
open NXyn.Aux.OtherMisc
open NXyn.Aux.SetMisc

open Basics
open LocalScheduling







let liveness (b:IBlock) (inout0:Set<Operand>) (dead0:Set<Operand>) (instrs:Instr []) =
             (instrs, ([],dead0))
             ||> Array.foldBack (fun i (augs,v_dead) -> let v_d    = D_of_Instr                              i |> Set.filter Operand_is_Vector
                                                        let v_ds   = DS_of_Instr                             i |> Set.filter Operand_is_Vector
                                                        let v_s    = S_of_Instr  konstSizeDict b.ReturnsVoid i |> Set.filter Operand_is_Vector

                                                        let v_dead = v_dead + v_d
                                                        let v_kill = Set.intersect v_s v_dead
                                                        let v_dead = v_dead - v_s
                                                        let v_in   = v_s + v_ds

                                                        let v_in   = v_in - inout0 
                                                        let v_kill = v_kill - inout0
                                                        let v_d    = v_d - inout0
                                                        let v_dead = v_dead - inout0
 
                                                        ((v_in,v_kill,v_d,i)::augs,v_dead))



(* => X64instr *)
//let Int32_of_R8(r:R8) = r.Int
//let Int32_of_R16(r:R16) = r.Int
//let Int32_of_R32(r:R32) = r.Int
//let Int32_of_R64(r:R64) = r.Int
//let Int32_of_X(x:X) = x.Int
//w
//let Get_S_of_Instr (kDict:Dictionary<String,Int32>) (returnsVoid:Boolean) (i:Instr) =
//          let (totalsize,prep) = List.fold_right prep_of_liveness augs (0,[])


type prep = (int [] * int [] * int [] * Instr * int [])

let prep_of_liveness (inout0:Set<Operand>) (augs:list<Set<Operand> * Set<Operand> * Set<Operand> * Instr>) : list<prep> =
                     let map1 (a_in,a_kill,a_gen,instr) = 
                              (* we only care about vector operands here. map vector registers to their register index *)
                              let v_in   = a_in   |> Set.filter Operand_is_Vector |> Set.toArray |> Array.map Int32_of_VectorOperand
                              let v_kill = a_kill |> Set.filter Operand_is_Vector |> Set.toArray |> Array.map Int32_of_VectorOperand
                              let v_gen  = a_gen  |> Set.filter Operand_is_Vector |> Set.toArray |> Array.map Int32_of_VectorOperand

                              (* generate all variations of this instruction *)
                              let lookupX ix ofs where what =
                                          if   Set.contains (Operand_of_X(what)) inout0                                 then what
                                          elif ix &&& (1 <<< (ofs + Array.findIndex ((=)(Int32_of_X(what))) where)) = 0 then X.Lo
                                                                                                                        else X.Hi
                              let lookupY ix ofs where what =
                                          if   Set.contains (Operand_of_Y(what)) inout0                                 then what
                                          elif ix &&& (1 <<< (ofs + Array.findIndex ((=)(Int32_of_Y(what))) where)) = 0 then Y.Lo
                                                                                                                        else Y.Hi
                              let variations = Array.init (1 <<< (v_in.Length + v_gen.Length))
                                                          (fun ix -> instr
                                                                     |> Map_Yds_and_Ys_in_Instr (lookupY ix 0 v_in) |> Map_Yd_in_Instr (lookupY ix v_in.Length v_gen)
                                                                     |> Map_Xds_and_Xs_in_Instr (lookupX ix 0 v_in) |> Map_Xd_in_Instr (lookupX ix v_in.Length v_gen))

                              (* return arrays of vector register indices, the instruction, and all of its variations modulo vector-register renaming *)
                              (v_in,v_kill,v_gen,instr,variations)

                     (* map liveness information using auxiliary function above *)
                     let info = List.map map1 augs

                     (* collect all instruction variations, obtain their respectives sizes *)
                     info |> List.map (fun (_,_,_,_,variations) -> variations) |> Array.concat |> InstrSize.sizes_of_Instrs08 |> ignore

                     (* `InstrSize.size_of_Instr' depends on GNU assembler. all instructions considered should be cached by now *) 
                     info |> List.map (fun (v_in,v_kill,v_gen,instr,variations) -> (v_in,v_kill,v_gen,instr,variations |> Array.map InstrSize.size_of_Instr))




//open Genes01

(* evaluate the fitness of some individual (without the intermediate step of generating actual instructions) *)
let evaluate_genes (size0:Int32) ((nVfreeLo0,nVfreeHi0):(Int32*Int32)) (p:prep []) (g:Genes01.t) =
                   (* initial mapping of vector registers: identity (lo -> lo, hi -> hi) *)
                   let mutable vMapping  = 0xff00

                   let mutable nVfreeLo  = nVfreeLo0
                   let mutable nVfreeHi  = nVfreeHi0

                   let mutable totalSize = size0
//                   let mutable nRepair   = 0

                   (* iterate over all genes of this individual *)
                   let mutable pos = 0
                   for i = 0 to p.Length - 1 do
                         let (v_in,v_kill,v_gen,_,sizes) = p.[i]

                         (* successively build up index `ix' into `sizes' array *)
                         let mutable ix = 0
                         for j = 0 to v_in.Length   - 1 do if vMapping &&& (1 <<< v_in.[j]) <> 0 then ix <- ix ||| (1 <<< j)
                         for j = 0 to v_kill.Length - 1 do if vMapping &&& (1 <<< v_kill.[j]) = 0 then nVfreeLo <- nVfreeLo + 1 else nVfreeHi <- nVfreeHi + 1
                         for j = 0 to v_gen.Length  - 1 do
                               let d = v_gen.[j]
                               
                               if g.[pos] then
                                  (* individual requests a register out of the upper half *)
                                  if   nVfreeHi > 0 then (* requests can be satisfied. decrease number of free registers in upper half and set bit in mapping. *)
                                                         nVfreeHi <- nVfreeHi - 1
                                                         vMapping <- vMapping ||| (1 <<< d)
                                                         ix <- ix ||| (1 <<< (j + v_in.Length))
                                  elif nVfreeLo > 0 then (* original request cannot be met, but repair is possible: flip this gene and take register from lower half *)
                                                         nVfreeLo <- nVfreeLo - 1
//                                                         nRepair <- nRepair + 1
//                                                         g.MutateClear1(pos)
                                                         vMapping <- vMapping &&& (~~~ (1 <<< d))
                                                    else (* request cannot be met at all *)
                                                         failwith "OptVecReallocGA.evaluate_genes (1)"
                               else (* individual requests a register out of the lower half *)
                                    if   nVfreeLo > 0 then (* request can be satisfied. decrease # of registers in lower half and clear bit in mapping. *)
                                                           nVfreeLo <- nVfreeLo - 1
                                                           vMapping <- vMapping &&& (~~~ (1 <<< d))
                                    elif nVfreeHi > 0 then (* original request cannot be met, but repair is possible: flip this gene and take register from upper half *)
                                                           nVfreeHi <- nVfreeHi - 1
//                                                           nRepair <- nRepair + 1
//                                                           g.MutateSet1(pos)
                                                           vMapping <- vMapping ||| (1 <<< d)
                                                           ix <- ix ||| (1 <<< (j + v_in.Length))
                                                      else (* request cannot be met at all *)
                                                           failwith "OptVecReallocGA.evaluate_genes (2)"

                               (* for each vector register allocated, advance by one in the genes array *)
                               pos <- pos + 1

                         (* add size of this instruction to total *)
                         totalSize <- totalSize + sizes.[ix]

                   float(-totalSize)



(* map genes to array of corresponding instructions *)
let express_genes (inout0:Set<Operand>) (vec_uninitialized:Set<Operand>) (p:prep []) (g:Genes01.t) =
                  let freeV    = (X.All |> List.map Operand_of_X |> Set.ofList) - vec_uninitialized - inout0
                  let mappingV = Array.init 16 (fun i -> i)
 
                  let v_freeLo = Set.filter (fun(v) -> Int32_of_VectorOperand v <  8) freeV |> Set.toList |> List.map Int32_of_VectorOperand |> ref
                  let v_freeHi = Set.filter (fun(v) -> Int32_of_VectorOperand v >= 8) freeV |> Set.toList |> List.map Int32_of_VectorOperand |> ref

                  let v_free(hi:Boolean) = if hi then if !v_freeHi <> [] then v_freeHi else v_freeLo
                                                 else if !v_freeLo <> [] then v_freeLo else v_freeHi
//                                                 if (not hi) && !v_freeLo <> [] then v_freeLo
//                                                                            else v_freeHi  

                  let map_X (v_map:(Int32 * Int32) []) (v:X) = if Set.contains (Operand_of_X v) inout0 then v else X128(snd (Array.find (fun (x,_) -> x = v.Int) v_map))
                  let map_Y (v_map:(Int32 * Int32) []) (v:Y) = if Set.contains (Operand_of_Y v) inout0 then v else Y256(snd (Array.find (fun (x,_) -> x = v.Int) v_map))

                  let pos = 0 |> ref
                  p |> Array.mapi (fun i (v_in,v_kill,v_gen,instr,_) ->
                                       let v_ds_map = v_in |> Array.map (fun (v:int) -> (v, mappingV.[v]))

                                       for j = 0 to v_kill.Length - 1 do let r = mappingV.[v_kill.[j]] in if r < 8 then v_freeLo := r::!v_freeLo else v_freeHi := r::!v_freeHi
   
                                       for j = 0 to v_gen.Length - 1 do
                                             let v_free = v_free(g.[!pos])
                                             incr(pos)
                                             match !v_free
                                              with | [] -> failwithf "express_genes %d" i
                                                   | hd::tl -> mappingV.[v_gen.[j]] <- hd
                                                               v_free := tl
 
                                       let v_d_map = v_gen |> Array.map (fun (v:int) -> (v, mappingV.[v]))
 
                                       instr |> InstrIO.Map_Xds_and_Xs_in_Instr (map_X v_ds_map) |> InstrIO.Map_Xd_in_Instr (map_X v_d_map)
                                             |> InstrIO.Map_Yds_and_Ys_in_Instr (map_Y v_ds_map) |> InstrIO.Map_Yd_in_Instr (map_Y v_d_map))


















let doit (how:ReassignXMM) (b:IBlock) ((in0,out0):(Set<Operand> * Set<Operand>)) (instrs:Instr []) =

    let inout0 = Set.union in0 out0


    if   instrs |> Array.exists (fun(i) -> FixedRs_of_Instr(i) <> [] || FixedRds_of_Instr(i) <> [] || FixedRd_of_Instr(i) <> []) then instrs
    elif instrs |> Array.exists (fun(i) -> List.exists (function Rbyte(Hi(_)) -> true | _ -> false) (Rd_of_Instr(i) @ Rds_of_Instr(i) @ Rs_of_Instr(i))) then instrs
    else
//let AllRs_of_Instr(i:Instr)  = FixedRs_of_Instr(i)  @ Rs_of_Instr(i)
//let AllRds_of_Instr(i:Instr) = List.filter (fun(r) -> size_of_R(r) <  4) (FixedRd_of_Instr(i) @ Rd_of_Instr(i)) @ FixedRds_of_Instr(i) @ Rds_of_Instr(i)
//let AllRd_of_Instr(i:Instr)  = List.filter (fun(r) -> size_of_R(r) >= 4) (FixedRd_of_Instr(i) @ Rd_of_Instr(i))

//      let inout0 = Set.union (List.map Operand_of_R R.All |> Set.ofList) inout0

//      InstrSize.PrepareForInstrs08(instrs)
      let size0 = instrs |> InstrSize.sizes_of_Instrs |> Array.sum // Array.map InstrSize.size_of_Instr |> Array.sum

//      let size0 = instrs |> Array.map InstrSize.size_of_Instr |> Array.sum

      eprintfn "GA: size0 = %d" size0

//      let all = (List.map Operand_of_X X.All |> Set.ofList) - inout0
      let all = (Set.union (List.map Operand_of_X X.All |> Set.ofList) (List.map Operand_of_R R.All |> Set.ofList)) - inout0
//      let i_all = (List.map Operand_of_R R.All |> Set.ofList) - inout0

      let (augs,vecdead) = liveness b inout0 all instrs
//      let (augs,other) = List.partition (fun (v_in,v_kill,v_gen,instr) -> Set.count(v_in) + Set.count(v_gen) > 0) augs

//      let other = Array.ofList (List.map (fun(_,_,_,i) -> i) other)
//      InstrSize.PrepareForInstrs08(other)
//      let othersize = 0 // other |> Array.map InstrSize.size_of_Instr |> Array.sum
          

          
      let vfree0 = (Set.union (List.map Operand_of_X X.All |> Set.ofList) (List.map Operand_of_R R.All |> Set.ofList)) - vecdead
//      let vfree0 = (List.map Operand_of_X X.All |> Set.ofList) - vecdead
//          let vec_uninitialized = (Set.of_list X.All) - vecdead


      let prep = prep_of_liveness inout0 augs
(*
      let prep = List.map (prep_of_liveness inout0) augs
//      let (totalsize,prep) = List.foldBack (prep_of_liveness inout0) augs (0,[])


      let allinstrs = Array.concat (List.map (fun (_,_,_,_,instrs) -> instrs) prep)
      allinstrs |> InstrSize.sizes_of_Instrs08 |> ignore
      

      let prep = List.map (fun (v_in,v_kill,v_gen,instr,instrs) -> let sizes = Array.map InstrSize.size_of_Instr instrs
                                                                   (v_in,v_kill,v_gen,instr,sizes)) prep
*)
//      let prep = List.map (fun (i_in,i_kill,i_gen,v_in,v_kill,v_gen,instr,instrs) -> let sizes = Array.map InstrSize.size_of_Instr instrs
//                                                                                     (i_in,i_kill,i_gen,v_in,v_kill,v_gen,instr,sizes)) prep


      let nbits = List.sum (List.map (fun (_,_,v_gen,_,_) -> Array.length v_gen) prep)

//      Printf.printf "totalsize = %d (hoisted)\n" totalsize
      Printf.printf "nbits   = %d\n" nbits
//          Printf.printf "eval maxcosts = %d\n" maxcosts
//          Printf.printf "eval asis = %d (%d + %d)\n" all_size totalsize (all_size - totalsize)
      if nbits = 0 then instrs
      else
          

          let freeV = (X.All |> List.map Operand_of_X |> Set.ofList) - vfree0 - inout0
          let v_freeLo = Set.filter (fun(v) -> Int32_of_VectorOperand v <  8) freeV |> Set.toList |> List.map Int32_of_VectorOperand
          let v_freeHi = Set.filter (fun(v) -> Int32_of_VectorOperand v >= 8) freeV |> Set.toList |> List.map Int32_of_VectorOperand
          let nVfreeLo = List.length(v_freeLo)
          let nVfreeHi = List.length(v_freeHi)
          let nfree0 = (nVfreeLo, nVfreeHi)



          let prep = Array.ofList prep

          let (fixedV,prepV) = prep |> Array.partition (fun (v_in,v_kill,v_gen,_,sizes) -> Array.length(sizes) = 1)
          let hoistedcosts = Array.sum (Array.map (fun(_,_,_,_,sizes) -> (sizes:Int32 []).[0]) fixedV)

          let fitness_of_genes = evaluate_genes hoistedcosts nfree0 (* inout0 vfree0 *) prepV 


// hillclimbing
          let hillclimb ((fitness0,best0):(float * Genes01.t)) =
            let genesBEST   = ref best0
            let fitnessBEST = ref fitness0
            
            for i = 0 to nbits - 1 do
                let genes1 = (!genesBEST).DeepCopy()
                genes1.MutateFlip1(i)

                let fitness1 = fitness_of_genes(genes1) // eval_genes hoistedcosts nfree0 prepV genes1 //

                if fitness1 > !fitnessBEST then
//                    eprintfn "HC new best %f" fitness1
                    fitnessBEST := fitness1
                    genesBEST   := genes1
            (!fitnessBEST,!genesBEST)



          //



          let all0 = Genes01.t.Replicate( 0L,nbits)
          let all1 = Genes01.t.Replicate(-1L,nbits)

//          eprintfn "othersize = %d" othersize


          (* lower and upper bounds *)
          let instrs_all_lo = instrs |> Array.map (Map_X_in_Instr (fun _ -> X.Lo) >> Map_Y_in_Instr (fun _ -> Y.Lo))
          let instrs_all_hi = instrs |> Array.map (Map_X_in_Instr (fun _ -> X.Hi) >> Map_Y_in_Instr (fun _ -> Y.Hi))

          let size_all_lo = instrs_all_lo |> InstrSize.sizes_of_Instrs |> Array.sum
          let size_all_hi = instrs_all_hi |> InstrSize.sizes_of_Instrs |> Array.sum


          // 


          let costs0 = fitness_of_genes(all0)
          let costs1 = fitness_of_genes(all1)
//          let costs0 = costs0 + othersize
//          let costs1 = costs1 + othersize

          eprintfn "size_all_lo = %d" size_all_lo
          
          eprintfn "eval(all0) = %A" costs0
          eprintfn "eval(all1) = %A" costs1


          let f_all0 = costs0
          let g_all0 = all0
          let (hc_f0,hc_b0) = hillclimb (f_all0,g_all0)

          let f_all1 = costs1
          let g_all1 = all1
          
          let (hc_f1,hc_b1) = hillclimb (f_all1,g_all1)


          eprintfn "size_all_hi = %d" size_all_hi


          eprintfn "fitness hc all0 : %A => %A" f_all0 hc_f0
          eprintfn "fitness hc all1 : %A => %A" f_all1 hc_f1

          let doGA = match how with ReassignXMM.No | ReassignXMM.HC -> false | ReassignXMM.GA -> true
          if not(doGA) then
              
             (* return hc(all0) *)
             (hc_f0,hc_b0)

          else let rng = new System.Random()
               let best = ref (min costs0 costs1)
               let pop0 = new ResizeArray<Genes01.t>()

               for i = 0 to 1000 do pop0.Add(Genes01.t.Random(rng,nbits))
       
               for i = 0 to 1000 do pop0.Add(Genes01.t.Random(rng,nbits))
               let pop0 = pop0.ToArray()

               let ga = new GA.t<Genes01.t>(Evaluate          = fitness_of_genes, // eval_genes hoistedcosts nfree0 prepV, // (eval_genes nfree0 prep >> fun(costs,_,_,_,_) -> -(float costs)),
                                            Crossover         = (fun (xx:Genes01.t) (yy:Genes01.t) -> let (xy,yx) = xx.CrossoverMulti([| rng.Next(nbits); rng.Next(nbits) |],yy) in [| xy; yx |]),
                                            Generate          = (fun () -> Genes01.t.Random(rng,nbits)),
                                            Mutate            = (fun(x:Genes01.t) -> x.MutateFlip1(rng.Next(x.Size))),
                                            MutateProbability = 0.0001,
                                            Verbosity         = 100,
                                            PopulationSize    = 2500,
                                            RNG               = rng,
                                            Selection         = GA.TournamentSelection(0.7,2))

               let allBEST = ga.Do(1000, pop0, 50)

//               let (fitnessBEST,genesBEST) = allBEST.[0]
//               eprintfn "===>  best = %A, costsBEST = %f" genesBEST fitnessBEST

               //
               allBEST
               |> Array.map (fun(x) -> async { return(hillclimb(x)) })
               |> Async.Parallel
               |> Async.RunSynchronously
               |> Array.maxBy fst

          
          |> fun (fitnessBEST,genesBEST) ->


                 let instrsBEST = express_genes inout0 vfree0 prep genesBEST
                 let size1 = instrsBEST |> InstrSize.sizes_of_Instrs |> Array.sum

                 eprintfn "GA: fitness = %A, size1 = %d" fitnessBEST size1

                 instrsBEST // b.Instrs <- instrsBEST
