#light (* -*- tuareg -*- *)

module Invoke_as

let psi = new System.Diagnostics.ProcessStartInfo("as", "--64 -a -o /dev/null")
psi.CreateNoWindow <- true
psi.RedirectStandardInput <- true
psi.RedirectStandardOutput <- true
psi.RedirectStandardError <- true
psi.UseShellExecute <- false




(* string list -> int list *)
let nopLine = ".byte 0x090"

let codeSize_of_instrString (asmLines:string array) =
  let n = Array.length asmLines

  let asmLengths = Array.create n 0
  use ps = new System.Diagnostics.Process()

  let lastofs = ref 0
  let m = ref 0

  let processReply src (outLine:System.Diagnostics.DataReceivedEventArgs) =
    let line = outLine.Data
    if not (System.String.IsNullOrEmpty(line)) then
      let split = line.Split([| ' '; '\t' |], System.StringSplitOptions.RemoveEmptyEntries)
      if Array.length split > 3 && split.[2] = "90" then 
        let ofs = System.Int32.Parse(split.[1], System.Globalization.NumberStyles.HexNumber)
        asmLengths.[!m] <- ofs - !lastofs
        incr m
        lastofs := ofs + 1

  ps.OutputDataReceived.AddHandler(new System.Diagnostics.DataReceivedEventHandler(processReply))
  ps.StartInfo <- psi
  ps.Start() |> ignore

  ps.BeginOutputReadLine()

  ps.StandardInput.WriteLine(".psize 0")
  for i = 0 to n - 1 do 
    ps.StandardInput.WriteLine(asmLines.[i])
    ps.StandardInput.WriteLine(nopLine)
  ps.StandardInput.Close()
  ps.WaitForExit()

  if !m = n then Some asmLengths else None


let main (args:string array) =
  Printf.printf "System.Environment.ProcessorCount = %d\n" System.Environment.ProcessorCount

  let lines = System.IO.File.ReadAllLines(args.[1])

  for i = 0 to 1000 do
    let sizes = codeSize_of_instrString lines
    ignore sizes

  let sizes = codeSize_of_instrString lines
  match sizes with
    | None -> failwith "FAILURE"
    | Some xs ->
        let sum = Array.fold (+) 0 xs
(*        for i = 0 to Array.length xs - 1 do
          Printf.printf "len(%s) = %d\n" lines.[i] xs.[i]
*)
        Printf.printf "\n\ntotal length = %x\n" sum

main (System.Environment.GetCommandLineArgs ())

