module NXas_parse

open System
open System.Collections.Generic

open Ast
open X64instr

open NXyn.Aux

open NXyn.Aux.Global

(***)


(* => global *)
(*
let g_inFileName  = ref ""
let g_outFileName = ref ""
let g_logFileName = ref ""


let g_appName = ref ""
*)
(* => global *)
(*
let Fail(s:String) = ProcessMisc.append_line_to_file (sprintf "[NXyn error] module=%s, inFile=%s, error=%s" !g_appName !g_inFileName s) !g_logFileName |> ignore
                     exit(2)

let Dump(s:String) = ProcessMisc.append_line_to_file (sprintf "[NXyn info] module=%s, inFile=%s: info=%s" !g_appName !g_inFileName s) !g_logFileName |> ignore

*)


open AstIO
open MoreAstIO

let sliceSections lines = let allSections = new ResizeArray<String * Item list * array<ELine>>()
                          let beginSection name args = (name,args,new ResizeArray<ELine>())
                          let closeSection(name,args,lines) = allSections.Add(name,args,(lines:ResizeArray<ELine>).ToArray()) 
                          let doLine s0 = function Tagged(".section",Iexpr(Id({Name = ((".data"|".text"|".bss") as name)}))::_) -> closeSection s0; beginSection name []
                                                 | Tagged(".section",CompactString(s)::args) -> closeSection s0; beginSection s args
                                                 | Tagged(".section",args) -> LogError(sprintf "sliceSections: unknown arguments '%A'" args)
                                                 | Tagged((".data"|".text"|".bss") as name,_) -> closeSection s0; beginSection name []
                                                 | line -> let (_,_,lines) = s0 in lines.Add(line); s0 
                          Array.fold doLine (beginSection ".text" []) lines |> closeSection
                          allSections.ToArray()
(***)

let nErrors = ref 0

let main (logFile:String) (inFileName:String) (rlines:array<ELine>) =
    (* assure no indirect calls/branches are used *)
    if rlines |> Array.exists (fun(r:ELine) -> r.UsesIndirectAddressing) then LogError("indirect branches not supported")

    (* extract sections from lines *)
    rlines
    |> sliceSections

    (* eliminate DWARF related sections *)
    |> Array.filter (fun(name,_,_) -> name <> ".eh_frame" && name <> ".debug_line")

    (* join identical section parts (e.g., all ".text" into one) *)
    |> fun(sections) -> let sDict = new Dictionary<string * Item list, ResizeArray<ELine []>>()
                        sections |> Array.iter (fun(name,args,lines) -> let (found,arrs) = sDict.TryGetValue((name,args))
                                                                        if found then arrs.Add(lines)
                                                                                 else let a = new ResizeArray<ELine []>() in a.Add(lines); sDict.[(name,args)] <- a)
                        ArrayMisc.array_of_collection(sDict)
                        |> Array.map(fun(KeyValue((name,args),arrs)) -> (name,args,arrs.ToArray() |> Array.concat))

    (* assure there is exactly one ".text" section with actual code in it *)
    |> fun(sections) -> match Array.filter (fun(name,_,_) -> name = ".text") sections
                         with | [|(_,_,lines)|] when lines.Length > 0 -> ()
                              | x -> LogError(sprintf "expected exactly one text section without any subsections %A" x)
                        sections

    (* process input lines: extract instructions/directives from tagged. push alignment into labels. *)
    |> Array.map (fun(name,args,lines0) -> let lines = new ResizeArray<ILine>()
                                           let doLine (align:Int32) (rline:ELine) =
                                                      match ILine_of_ELine(rline)
                                                       with | None -> incr nErrors; LogInfo(sprintf "unknown asm line %A" rline); 1
                                                            | Some(IgnoredDirective(_)) -> align
                                                            | Some(UnsupportedDirective(d)) -> incr nErrors; LogInfo(sprintf "unsupported directive: %s" d); align
                                                            | Some(AlignmentDirective(n)) -> max n align
                                                            | Some(ALabel(name,n)) -> lines.Add(ALabel(name,max n align)); 1
                                                            | Some(line) -> lines.Add(line); align
                                           Array.fold doLine 1 lines0 |> ignore
                                           (name,args,lines.ToArray()))
    |> if !nErrors = 0 then id else LogError(sprintf "%d input lines were unknown or not supported" !nErrors)

(***)

(* BOILER? *)
Environment.GetCommandLineArgs()
|> function [| _; logFile; inFile; outFile |] -> Global.InFileName  := inFile
                                                 Global.OutFileName := outFile
                                                 Global.LogFileName := logFile
                                                 Global.AppName     := "nxas_parse"
                                                 use r = new IO.StreamReader(inFile)
                                                 let lexbuf = Microsoft.FSharp.Text.Lexing.LexBuffer<char>.FromTextReader(r)
                                                 let inLines = try  AttParser.ASTlines AttLexer.token lexbuf 
                                                               with exn -> LogError(sprintf "lexer error near character %d: %A" lexbuf.EndPos.Column exn)
                                                 r.Close()
                                                 let outProgram = main logFile inFile inLines
                                                 use w = new IO.StreamWriter(outFile)
                                                 JSONvalue_of_Program(outProgram).Write(w)
                                                 w.Close()
          | _ -> eprintfn "usage: nxas_parse <logfile> <input.S> <output.nxas.JSON>"
                 exit(3)
